from Piece import Piece, PieceType, PieceColor, pieceTypeFromRepr
import numpy as np
import copy
from functools import cache #TODO: caching

class Null:
    def __init__(self):
        self.color = PieceColor.EMPTY
        self.type = PieceType.EMPTY

    def __repr__(self):
        return '·'

    def __eq__(self, other):
        return other is None

    def __bool__(self):
        return False

    def isNull(self):
        return True

class GameState():
    def __init__(self, boardState=None, colorToMove=PieceColor.WHITE, singlePlayer=True):
        self.singlePlayer = singlePlayer
        self.white_pieces = []
        self.black_pieces = []
        self.colorToMove = colorToMove
        if boardState is None:
            self.board = self.__setup_new_board()
        else:
            self.board = boardState
            #populate piece lists
            for i in range(8):
                for j in range(8):
                    if boardState[i][j].color == PieceColor.WHITE:
                        self.white_pieces.append(boardState[i][j])
                    elif boardState[i][j].color == PieceColor.BLACK:
                        self.black_pieces.append(boardState[i][j])
        self.last_move = None

    def available_moves(self):
        if self.game_over():
            return []
        colorToMove = self.colorToMove
        moves = []
        assert colorToMove == PieceColor.WHITE or colorToMove == PieceColor.BLACK
        corr_piece_list = None
        if colorToMove == PieceColor.WHITE:
            corr_piece_list = self.white_pieces
        else:
            corr_piece_list = self.black_pieces

        for piece in corr_piece_list:
            if piece.taken: #sanity check
                continue
            if piece.type == PieceType.PAWN:
                moves += self.__available_pawn_moves(piece)
            elif piece.type == PieceType.ROOK:
                moves += self.__available_rook_moves(piece)
            elif piece.type == PieceType.KNIGHT:
                moves += self.__available_knight_moves(piece)
            elif piece.type == PieceType.BISHOP:
                moves += self.__available_bishop_moves(piece)
            elif piece.type == PieceType.QUEEN:
                #should be no overlap
                moves += self.__available_bishop_moves(piece)
                moves += self.__available_rook_moves(piece)
            elif piece.type == PieceType.KING:
                moves += self.__available_king_moves(piece)

        return sorted(self.__filter_forced_take(moves, colorToMove))

    def __filter_forced_take(self, moves, colorToMove):
        force_take_moves = []
        oppColor = PieceColor.BLACK if colorToMove == PieceColor.WHITE else PieceColor.WHITE
        for move in moves:
            endX, endY = self.__square_to_coords(move[2:4])
            if self.board[endX][endY].color == oppColor:
                force_take_moves.append(move)

        if len(force_take_moves) == 0:
            return moves
        return force_take_moves

    def __available_pawn_moves(self, piece):
        available = []
        if self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + 1))
        if piece.y - 1 > -1 and self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y - 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y - 1))
        if self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + 2)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + 2))
        if piece.y - 2 > -1 and  self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y - 2)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y - 2))
        if self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y + 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y + 1))
        if piece.x - 1 > -1 and self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y + 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y + 1))
        if piece.y - 1 > -1 and self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y - 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y - 1))
        if piece.x - 1 > -1 and piece.y - 1 > -1 and self.__is_valid_move(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y - 1)):
            available.append(self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y - 1))
        available += self.__available_pawn_promotions(piece)
        return available

    def __available_pawn_promotions(self, piece):
        av = []
        if piece.color == PieceColor.WHITE and piece.y == 7:
            if piece.y == 7: #already at edge (should never happen, but good check)
                return av
            if self.board[piece.x][piece.y + 1].isNull():
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")
            if piece.x - 1 > -1 and self.board[piece.x - 1][piece.y + 1].color == PieceColor.BLACK:
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y + 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")
            if piece.x + 1 < 8 and self.board[piece.x + 1][piece.y + 1].color == PieceColor.BLACK:
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y + 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")
        elif piece.color == PieceColor.BLACK and piece.y == 1:
            if piece.y == 0: #already at edge (should never happen, but good check)
                return av
            if self.board[piece.x][piece.y - 1].isNull():
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y - 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")
            if piece.x - 1 > -1 and self.board[piece.x - 1][piece.y - 1].color == PieceColor.WHITE:
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x - 1, piece.y - 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")
            if piece.x + 1 < 8 and self.board[piece.x + 1][piece.y - 1].color == PieceColor.WHITE:
                base = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + 1, piece.y - 1)
                av.append(base + "q")
                av.append(base + "n")
                av.append(base + "b")
                av.append(base + "r")

        return av

    def __available_rook_moves(self, piece):
        available = []
        vert = 1
        while piece.y + vert < 8:
            tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + vert)
            if self.__is_valid_move(tent_move):
                available.append(tent_move)
            vert += 1
        vert = -1
        while piece.y + vert > -1:
            tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x, piece.y + vert)
            if self.__is_valid_move(tent_move):
                available.append(tent_move)
            vert -= 1
        horz = 1
        while piece.x + horz < 8:
            tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + horz, piece.y)
            if self.__is_valid_move(tent_move):
                available.append(tent_move)
            horz += 1
        horz = -1
        while piece.x + horz > -1:
            tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + horz, piece.y)
            if self.__is_valid_move(tent_move):
                available.append(tent_move)
            horz -= 1
        return available

    def __available_knight_moves(self, piece):
        available = []
        possible_offsets = {(1,2), (1,-2), (-1,2), (-1,-2), (2,1), (2,-1), (-2,1), (-2,-1)}
        for i, j in possible_offsets:
            if piece.x + i < 8 and piece.x + i > -1 and piece.y + j < 8 and piece.y + j > -1:
                tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + i, piece.y + j)
                if self.__is_valid_move(tent_move):
                    available.append(tent_move)
        return available

    def __available_bishop_moves(self, piece):
        available = []
        possible_unit_offsets = {(1,1), (-1,1), (-1,-1), (1,-1)}
        for unit_offset in possible_unit_offsets:
            multiplicative_factor = 1
            while piece.x + unit_offset[0]*multiplicative_factor < 8 and piece.x + unit_offset[0]*multiplicative_factor > -1 and piece.y + unit_offset[1]*multiplicative_factor < 8 and piece.y + unit_offset[1]*multiplicative_factor > -1:
                tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + unit_offset[0]*multiplicative_factor, piece.y + unit_offset[1]*multiplicative_factor)
                if self.__is_valid_move(tent_move):
                    available.append(tent_move)
                multiplicative_factor += 1

        return available

    def __available_king_moves(self, piece):
        available = []
        possible_offsets = {(0,1), (1,0), (-1,0), (0,-1), (1,1), (-1,-1), (-1,1), (1,-1)}
        for i, j in possible_offsets:
            if piece.x + i < 0 or piece.x + i > 7 or piece.y + j < 0 or piece.y + j > 8:
                continue
            tent_move = self.__coords_to_square(piece.x, piece.y) + self.__coords_to_square(piece.x + i, piece.y + j)
            if self.__is_valid_move(tent_move):
                available.append(tent_move)

        return available


    def print_board(self):
        print(np.flip(self.board.transpose(), axis=0))

    def __setup_new_board(self):
        b = np.full((8, 8), Null(), dtype=object)

        for i in range(8):
            p = Piece(PieceType.PAWN, PieceColor.BLACK, i, 6)
            self.black_pieces.append(p)
            b[i][6] = p

        for i in range(8):
            p = Piece(PieceType.PAWN, PieceColor.WHITE, i, 1)
            self.white_pieces.append(p)
            b[i][1] = p

        piece_positions = [
            (PieceType.ROOK, PieceColor.BLACK, 0, 7), (PieceType.ROOK, PieceColor.BLACK, 7, 7),
            (PieceType.ROOK, PieceColor.WHITE, 0, 0), (PieceType.ROOK, PieceColor.WHITE, 7, 0),
            (PieceType.KNIGHT, PieceColor.BLACK, 1, 7), (PieceType.KNIGHT, PieceColor.BLACK, 6, 7),
            (PieceType.KNIGHT, PieceColor.WHITE, 1, 0), (PieceType.KNIGHT, PieceColor.WHITE, 6, 0),
            (PieceType.BISHOP, PieceColor.BLACK, 2, 7), (PieceType.BISHOP, PieceColor.BLACK, 5, 7),
            (PieceType.BISHOP, PieceColor.WHITE, 2, 0), (PieceType.BISHOP, PieceColor.WHITE, 5, 0),
            (PieceType.KING, PieceColor.BLACK, 4, 7), (PieceType.KING, PieceColor.WHITE, 4, 0),
            (PieceType.QUEEN, PieceColor.BLACK, 3, 7), (PieceType.QUEEN, PieceColor.WHITE, 3, 0)
        ]

        for piece_type, color, x, y in piece_positions:
            piece = Piece(piece_type, color, x, y)
            if color == PieceColor.WHITE:
                self.white_pieces.append(piece)
            else:
                self.black_pieces.append(piece)
            b[x][y] = piece

        return b

    def __square_to_coords(self, square):
        col = ord(square[0]) - ord('a')
        row = int(square[1]) - 1
        return col, row  # FLIPPED & zero indexed

    def __coords_to_square(self, x, y):
        square = chr(x + ord('a')) + str(y + 1)
        return square

    def __parse_uci_move(self, uci_move):
        start_square = uci_move[:2]
        startX, startY = self.__square_to_coords(start_square)
        end_square = uci_move[2:4]
        endX, endY = self.__square_to_coords(end_square)
        promoted_piece = uci_move[4] if len(uci_move) == 5 else None
        return startX, startY, endX, endY, promoted_piece

    def __is_castling(self, uci_move):
        startX, startY, endX, endY, promoted_piece = self.__parse_uci_move(uci_move)
        startKing = bool(self.board[startX][startY].type == PieceType.KING)
        return uci_move in {"e1g1", "e1c1", "e8g8", "e8c8"} and startKing

    def __path_not_obstructed(self, startX, startY, endX, endY):
        #checks only straight line paths
        if startX == endX:
            step = 0
            if startY < endY:
                step = 1
            else:
                step = -1
            multiplicative_factor = 1
            while startY + step*multiplicative_factor != endY:
                if not self.board[startX][startY + step*multiplicative_factor].isNull():
                    return False
                multiplicative_factor += 1
        elif startY == endY:
            step = 0
            if startX < endX:
                step = 1
            else:
                step = -1
            multiplicative_factor = 1
            while startX + step*multiplicative_factor != endX:
                if not self.board[startX + step * multiplicative_factor][startY].isNull():
                    return False
                multiplicative_factor += 1
        elif abs(startX - endX) == abs(startY - endY):
            x_step = 1 if startX < endX else -1
            y_step = 1 if startY < endY else -1
            for i in range(1, abs(startX - endX)):
                if not self.board[startX + i * x_step][startY + i * y_step].isNull():
                    return False
        else:
            return False #neither straight nor diagonal movement
        return True

    def __obeys_piece_movement_rules(self, piece, endX, endY):
        if piece.type == PieceType.KING:
            return (abs(piece.x - endX) == 0 and abs(piece.y - endY) == 1) ^ (abs(piece.x - endX) == 1 and abs(piece.y - endY) == 0)
        elif piece.type == PieceType.ROOK and self.__path_not_obstructed(piece.x, piece.y, endX, endY):
            return (abs(piece.x - endX) == 0 and abs(piece.y - endY) != 0) ^ (abs(piece.x - endX) != 0 and abs(piece.y - endY) == 0)
        elif piece.type == PieceType.BISHOP and self.__path_not_obstructed(piece.x, piece.y, endX, endY):
            return abs(piece.x - endX) == abs(piece.y - endY)
        elif piece.type == PieceType.QUEEN and self.__path_not_obstructed(piece.x, piece.y, endX, endY):
            return abs(piece.x - endX) == abs(piece.y - endY) or (abs(piece.x - endX) == 0 and abs(piece.y - endY) != 0) or (abs(piece.x - endX) != 0 and abs(piece.y - endY) == 0)
        elif piece.type == PieceType.KNIGHT: #can jump, so no obstructed path call
            return (abs(piece.x - endX) == 2 and abs(piece.y - endY) == 1) ^ (abs(piece.x - endX) == 1 and abs(piece.y - endY) == 2)
        elif piece.type == PieceType.PAWN:
            #does not yet support en passant
            correct_direction = bool((piece.color == PieceColor.WHITE and piece.y < endY) or (piece.color == PieceColor.BLACK and piece.y > endY))
            firstMove = bool((piece.y == 1 and piece.color == PieceColor.WHITE) or (piece.y == 6 and piece.color == PieceColor.BLACK))
            validDeltaY = bool((firstMove and (abs(piece.y - endY) == 2 or abs(piece.y - endY) == 1)) or (not firstMove and abs(piece.y - endY) == 1))
            taking = bool((not isinstance(self.board[endX][endY], Null)) and self.board[endX][endY].color != piece.color)
            validDeltaX = bool((taking and abs(piece.x - endX) == 1) or (not taking and abs(piece.x - endX) == 0))
            return correct_direction and validDeltaX and validDeltaY

    def __is_valid_move(self, uci_move):
        startX, startY, endX, endY, promoted_piece = self.__parse_uci_move(uci_move)
        if startX == endX and startY == endY: #must move
            return False
        if startX < 0 or startY < 0 or endX < 0 or endY < 0 or startX > 7 or startY > 7 or endX > 7 or endY > 7:
            return False
        if self.board[startX][startY].isNull():
            return False
        if self.board[startX][startY].color == self.board[endX][endY].color:
            return False
        if promoted_piece is not None and (promoted_piece == "p" or (self.board[startX][startY].type != PieceType.PAWN)):  # must promote pawn and cannot promote any piece that is not pawn
            return False
        if promoted_piece is not None and ((endY == 7 and self.board[startX][startY].color != PieceColor.WHITE) or (endY != 0 and self.board[startX][startY].color != PieceColor.BLACK)):  # tries to promote pawn when not allowed
            return False
        if self.__is_castling(uci_move): #castling not supported in antichess
            return False
        return self.__obeys_piece_movement_rules(self.board[startX][startY], endX, endY)

    def do_move(self, uci_move, inplace=False):
        if not self.singlePlayer:
            inplace=True
        startX, startY, endX, endY, promoted_piece = self.__parse_uci_move(uci_move)
        if not uci_move in self.available_moves():
            raise ValueError(f"{uci_move} is not a valid move.")
        if inplace:
            if promoted_piece is not None:
                self.board[startX][startY].x = endX
                self.board[startX][startY].y = endY
                self.board[endX][endY] = self.board[startX][startY]
                self.board[endX][endY].type = pieceTypeFromRepr(promoted_piece)
            else:
                if self.board[endX][endY].color != PieceColor.EMPTY:
                    self.__take(endX, endY)
                self.board[startX][startY].x = endX
                self.board[startX][startY].y = endY
                self.board[endX][endY] = self.board[startX][startY]
                self.board[startX][startY] = Null()
            self.toggleColor()
            self.last_move = uci_move
        else:
            gs = GameState(boardState=copy.deepcopy(self.board), colorToMove=self.colorToMove)
            gs.do_move(uci_move, inplace=True)
            gs.toggleColor()
            gs.last_move = uci_move
            return gs

    def toggleColor(self):
        self.colorToMove = PieceColor.WHITE if self.colorToMove == PieceColor.BLACK else PieceColor.BLACK

    def __take(self, x, y):
        if self.board[x][y].color == PieceColor.WHITE:
            self.white_pieces.remove(self.board[x][y])
        else:
            self.black_pieces.remove(self.board[x][y])
        self.board[x][y].set_taken(True)
        self.board[x][y] = Null()

    def black_score(self):
        score = 0
        for piece in self.black_pieces:
            if not piece.taken:
                score += piece.value()
        return score

    def white_score(self):
        score = 0
        for piece in self.white_pieces:
            if not piece.taken:
                score += piece.value()
        return score

    def game_over(self):
        return self.white_score() == 0 or self.black_score() == 0

    def result(self, color): # for mcts
        if (color == PieceColor.WHITE and self.white_score() == 0) or (color == PieceColor.BLACK and self.black_score() == 0):
            return 1
        if (color == PieceColor.WHITE and self.black_score() == 0) or (color == PieceColor.BLACK and self.white_score() == 0):
            return -1


import numpy as np
from GameState import GameState, Null
from Piece import Piece, PieceType, PieceColor, pieceTypeFromRepr

class MCTSNode:
    def __init__(self, state, parent=None, depth=0):
        self.state = state
        self.parent = parent
        self.children = []
        self.visits = 0
        self.wins = 0
        self.untried_moves = state.available_moves()
        self.depth = depth

    def is_fully_expanded(self):
        return len(self.untried_moves) == 0

    def best_child(self, c_param=1.4):
        choices_weights = [
            (child.wins / child.visits) + c_param * np.sqrt((2 * np.log(self.visits) / child.visits))
            for child in self.children
        ]
        return self.children[np.argmax(choices_weights)]

    def expand(self):
        move = self.untried_moves.pop()
        next_state = self.state.do_move(move)
        child_node = MCTSNode(next_state, parent=self, depth=self.depth + 1)
        self.children.append(child_node)
        return child_node

    def update(self, result):
        self.visits += 1
        self.wins += result


class Agent:
    def __init__(self, color, strategy=None):
        assert color == PieceColor.WHITE or color == PieceColor.BLACK, "Color must be PieceColor.WHITE or PieceColor.BLACK."
        self.color = color
        valid_strats = ["random",
                        "bfs-depth=2",
                        "bfs-depth=3",
                        "bfs-depth=2-e=.1",
                        "bfs-depth=3-e=.1",
                        "mcts-sims=1000-depth=10",
                        "mcts-sims=1000-depth=15",
                        "mcts-sims=1500-depth=10",
                        "minimax-depth=3",
                        "minimax-depth=5"]
        if strategy is None:
            print("Agent strategy not selected, defaulting to 'random'.")
            self.strategy = "random"
        elif not strategy in valid_strats:
            print(f"Specified strategy is not valid. Valid strategies are {valid_strats}.")
        else:
            self.strategy = strategy
            colorStr = "black" if self.color == PieceColor.BLACK else "white"
            print(f"Agent playing as {colorStr}, and using strategy {self.strategy}.")

    def get_move(self, gs):
        if self.strategy == "random":
            return self.random_move(gs)
        elif self.strategy == "bfs-depth=2":
            move_scores_dict = self.move_scores(gs, 2, self.__game_score, np.max)
            best_move = max(move_scores_dict, key=move_scores_dict.get)
            return best_move
        elif self.strategy == "bfs-depth=3":
            move_scores_dict = self.move_scores(gs, 3, self.__game_score, np.max)
            best_move = max(move_scores_dict, key=move_scores_dict.get)
            return best_move
        elif self.strategy == "bfs-depth=2-e=.1":
            move_scores_dict = self.move_scores(gs, 2, self.__game_score, np.max)
            best_move = max(move_scores_dict, key=move_scores_dict.get)
            if np.random.random() < .1:
                return np.random.choice(list(move_scores_dict.keys()))
            else:
                return best_move
        elif self.strategy == "bfs-depth=3-e=.1":
            move_scores_dict = self.move_scores(gs, 3, self.__game_score, np.max)
            best_move = max(move_scores_dict, key=move_scores_dict.get)
            if np.random.random() < .1:
                return np.random.choice(list(move_scores_dict.keys()))
            else:
                return best_move
        elif self.strategy == "mcts-sims=1000-depth=10":
            return self.mcts_move(gs, num_simulations=1000, max_depth=10)
        elif self.strategy == "mcts-sims=1000-depth=15":
            return self.mcts_move(gs, num_simulations=1000, max_depth=15)
        elif self.strategy == "mcts-sims=1500-depth=10":
            return self.mcts_move(gs, num_simulations=1500, max_depth=10)
        elif self.strategy == "minimax-depth=3":
            return self.minimax(gs, 3, -np.inf, np.inf, True)[1]
        elif self.strategy == "minimax-depth=5":
            return self.minimax(gs, 5, -np.inf, np.inf, True)[1]

    def random_move(self, gs):
        return np.random.choice(gs.available_moves())

    def move_scores(self, gs, depth, heuristic, metric):
        av_moves = gs.available_moves()
        if len(av_moves) == 0:
            return {}

        if depth == 0:
            return {move: heuristic(gs) for move in gs.available_moves()}

        move_scores_dict = {}
        for move in gs.available_moves():
            next_gs = gs.do_move(move)
            next_move_scores = self.move_scores(next_gs, depth - 1, heuristic, metric)
            if len(next_move_scores) != 0:
                move_scores_dict[move] = metric(list(next_move_scores.values()))
            else:
                move_scores_dict[move] = self.__game_score(next_gs)

        return move_scores_dict

    def __game_score(self, gs):
        if self.color == PieceColor.WHITE:
            return gs.black_score() - gs.white_score()
        else:
            return gs.white_score() - gs.black_score()

    def mcts_move(self, gs, num_simulations=1000, max_depth=15):
        root = MCTSNode(gs)

        for _ in range(num_simulations):
            node = root
            state = gs

            while node.is_fully_expanded() and node.children:
                node = node.best_child()
                state = node.state

            if not node.is_fully_expanded() and node.depth < max_depth:
                node = node.expand()
                state = node.state

            result = self.__mcts_simulate(state, max_depth - node.depth)

            while node:
                node.update(result)
                node = node.parent

        best_move = max(root.children, key=lambda c: c.visits).state.last_move
        return best_move

    def __mcts_simulate(self, state, remaining_depth):
        current_state = state
        depth = 0
        while not current_state.game_over() and depth < remaining_depth:
            current_state = current_state.do_move(np.random.choice(current_state.available_moves()))
            depth += 1
        if current_state.game_over():
            return 100 * current_state.result(self.color)
        else:
            return self.__game_score(current_state)

    def minimax(self, game_state, depth, alpha, beta, maximizing_player):
        if depth == 0 or game_state.game_over():
            return self.__game_score(game_state), None

        best_move = None
        if maximizing_player:
            max_eval = -np.inf
            for move in game_state.available_moves():
                new_game_state = game_state.do_move(move, inplace=False)
                eval, _ = self.minimax(new_game_state, depth - 1, alpha, beta, False)
                if eval > max_eval:
                    max_eval = eval
                    best_move = move
                alpha = max(alpha, eval)
                if beta <= alpha:
                    break
            return max_eval, best_move
        else:
            min_eval = np.inf
            for move in game_state.available_moves():
                new_game_state = game_state.do_move(move, inplace=False)
                eval, _ = self.minimax(new_game_state, depth - 1, alpha, beta, True)
                if eval < min_eval:
                    min_eval = eval
                    best_move = move
                beta = min(beta, eval)
                if beta <= alpha:
                    break
            return min_eval, best_move


from enum import Enum

class PieceType(Enum):
    PAWN = 1
    KNIGHT = 2
    BISHOP = 3
    ROOK = 5
    QUEEN = 6
    KING = 7
    EMPTY = 8

class PieceColor(Enum):
    WHITE = 1
    BLACK = 2
    EMPTY = 3

def pieceTypeFromRepr(r):
    dic = {
        "p" : PieceType.PAWN,
        "n" : PieceType.KNIGHT,
        "b" : PieceType.BISHOP,
        "r" : PieceType.ROOK,
        "q" : PieceType.QUEEN,
        "k" : PieceType.KING
    }
    return dic[r.lower()]

class Piece:
    def __init__(self, type=None, color=None, x=None, y=None):
        self.type = type
        self.color = color
        self.taken = False
        self.x = x
        self.y = y

    def value(self):
        dic = {
            PieceType.PAWN : 1,
            PieceType.KNIGHT : 3,
            PieceType.BISHOP : 3,
            PieceType.ROOK : 5,
            PieceType.QUEEN : 9,
            PieceType.KING : 3.5 #assuming king counts for 3.5 points
        }
        return dic[self.type]

    def isNull(self):
        return False

    def set_taken(self, t):
        if t:
            self.x = -1
            self.y = -1
        self.taken = t

    def __repr__(self):
        if self.taken:
            return "-"

        id = "p"
        if self.type == PieceType.KNIGHT:
            id = "n"
        elif self.type == PieceType.BISHOP:
            id = "b"
        elif self.type == PieceType.ROOK:
            id = "r"
        elif self.type == PieceType.QUEEN:
            id = "q"
        elif self.type == PieceType.KING:
            id = "k"

        if self.color == PieceColor.WHITE:
            return id.upper()
        return id

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        # place value encoding
        hsh = self.type.value * 10000 + self.color.value * 1000 + self.x * 100 + self.y * 10
        if not self.taken:
            hsh += 1

        return hsh

    def __eq__(self, other):
        return isinstance(other, Piece) and hash(self) == hash(other)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent
long_description = (here / "README.txt").read_text()

setup(
    name='antichess',
    version='0.1.1',
    description='Package for simulating and playing antichess games.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Divit Rawal',
    author_email='divit.rawal+antichess@gmail.com',
    url='https://github.com/divitr/antichess',
    packages=find_packages(),
    install_requires=[
        'numpy',
    ],
    python_requires='>=3.6',
)




import numpy as np
from GameState import GameState, Null
from Piece import Piece, PieceType, PieceColor, pieceTypeFromRepr
from Agent import Agent

class AntiChessGame:
    def __init__(self, boardState=None, colorToMove=None, singlePlayer=True, showRules=False):
        if boardState is not None and colorToMove is None:
            raise ValueError("If defining board state, color to move must also be defined.")
        if boardState is None and (colorToMove is not None or colorToMove != PieceColor.WHITE):
            print("Board state not defined. Initializing new board and defaulting color to move as white.")

        if colorToMove is None:
            colorToMove = PieceColor.WHITE

        self.showRules = showRules

        self.gs = GameState(boardState=boardState, colorToMove=colorToMove, singlePlayer=singlePlayer)

    def play(self):
        assert self.gs.singlePlayer == False, "Use playAgent() to play against an agent in single player mode."
        if self.showRules:
            print("""Rules:
                    1. All pieces move the same as they do in regular chess, with the exception of castling and en passant, neither of which is allowed.
                    2. Capturing the king in AntiChess does not end the game.
                    3. The winner is the first person to have all of their pieces captured.
                    4. If you can capture an opponent's piece during your turn, you must capture it.
                    5. Moves must be entered in UCI chess notation.
                    6. Typing 'moves' into the prompt will print a list of valid moves for the turn.
                    7. Typing 'quit' into the prompt will forfeit the game. By default, your opponent will win.

                    """)
        while True:
            if self.gs.game_over():
                winner = "black" if self.gs.black_score() == 0 else "white"
                print(f"Game Over! {winner} won.")
                break
            self.display_board()
            if self.__get_move():
                if self.gs.colorToMove == PieceColor.WHITE:
                    print("Black wins by forfeit.")
                    break
                else:
                    print("White wins by forfeit.")
                    break

    def playAgent(self, strategy=None, agentColor=PieceColor.BLACK):
        assert self.gs.singlePlayer == True, "Use play() to play against an agent in two player mode."
        assert agentColor == PieceColor.WHITE or agentColor == PieceColor.BLACK, "Agent color must be black or white."
        agent = Agent(strategy=strategy, color=agentColor)
        agentcolorStr = "black" if agentColor == PieceColor.BLACK else "white"
        if self.showRules:
            print("""Rules:
                    1. All pieces move the same as they do in regular chess, with the exception of castling and en passant, neither of which is allowed.
                    2. Capturing the king in AntiChess does not end the game.
                    3. The winner is the first person to have all of their pieces captured.
                    4. If you can capture an opponent's piece during your turn, you must capture it.
                    5. Moves must be entered in UCI chess notation.
                    6. Typing 'moves' into the prompt will print a list of valid moves for the turn.
                    7. Typing 'quit' into the prompt will forfeit the game. By default, your opponent will win.

                    """)
        while True:
            if self.gs.game_over():
                winner = "black" if self.gs.black_score() == 0 else "white"
                print(f"Game Over! {winner} won.")
                break
            self.display_board()
            if self.gs.colorToMove == agentColor:
                print(f"{agentcolorStr} move > thinking...", end="", flush=True)
                uci_move = agent.get_move(self.gs)
                self.gs.do_move(uci_move, inplace=True)
                print(f"\r{agentcolorStr} move > {uci_move}                   ")
            elif self.__get_move():
                if self.gs.colorToMove == PieceColor.WHITE:
                    print("Black wins by forfeit.")
                    break
                else:
                    print("White wins by forfeit.")
                    break

    def display_board(self):
        board = np.flip(self.gs.board.transpose(), axis=0)
        print()
        for i in range(8, 0, -1):
            row_str = " ".join([str(x) for x in board[8-i]])
            print(f"{i}│ {row_str}")
        print("  ————————————————")
        print("   a b c d e f g h")
        print()

    def __get_move(self): # returns True if quitting
        while True:
            toMove = "white" if self.gs.colorToMove == PieceColor.WHITE else "black"
            try:
                move = input(f"{toMove} move > ")
                if move == "quit":
                    return True
                if move == "moves":
                    print(self.gs.available_moves())
                else:
                    self.gs.do_move(move, inplace=True)
                    break
            except ValueError as ve:
                print(ve)
            except Exception as e:
                print(f"Something went wrong: {e}")
                break